/* {=================================================================
 *
 * luapgplot.c
 * Lua interface to PGPLOT
 * Luis Carvalho (carvalho @ dam.brown.edu)
 * See Copyright Notice in luapgplot.h
 * $Id: $
 *
 * ==================================================================} */

#include <lua.h>
#include <lauxlib.h>
#include <stdlib.h>
#include <string.h>
#include "luapgplot.h"

/* default tr */
#define TRSIZE (6)
static float STD_TR[TRSIZE] = {0.0, 1.0, 0.0, 0.0, 0.0, 1.0};

#define TABLE_NONNUM_ERROR "table contains non-numeric value"
#define FUNCTION_NONNUM_ERROR "function returned non-numeric value"

/* Auxiliar */
static float pgplot_checkfloat (lua_State *L, int pos, int narg,
    const char *msg) {
  lua_Number d = lua_tonumber(L, pos);
  if (d == 0 && !lua_isnumber(L, pos)) luaL_argerror(L, narg, msg);
  return (float) d;
}

static float pgplot_optfloat (lua_State *L, int narg, float def) {
  lua_Number d = lua_tonumber(L, narg);
  if (d == 0 && !lua_isnumber(L, narg)) return def;
  return (float) d;
}

static void pgplot_avdevices (lua_State *L) {
  int i, n;
  char type[MAX_TYPE_SIZE];
  char descr[MAX_DESCR_SIZE];
  pgqndt_(&n); /* # available devices */
  lua_createtable(L, n, 0); /* result */
  for (i = 1; i <= n; i++) {
    int tlen, dlen, inter;
    pgqdt_(&i, type, &tlen, descr, &dlen, &inter,
        MAX_TYPE_SIZE, MAX_DESCR_SIZE);
    lua_createtable(L, 0, 3); /* i-th device */
    lua_pushlstring(L, type, (size_t) tlen);
    lua_setfield(L, -2, "type");
    lua_pushlstring(L, descr, (size_t) dlen);
    lua_setfield(L, -2, "description");
    lua_pushboolean(L, inter);
    lua_setfield(L, -2, "interactive");
    lua_rawseti(L, -2, i);
  }
}


/* {=================================================================
 *    float Buffer
 * ==================================================================} */

static float *table_fbuffer (lua_State *L, int pos, int n, float *data) {
  int i;
  /* traverse ipairs and write to p */
  for (i = 0; i < n; i++) {
    lua_rawgeti(L, pos, i + 1);
    data[i] = pgplot_checkfloat(L, -1, pos, TABLE_NONNUM_ERROR);
    lua_pop(L, 1);
  }
  return data;
}

static float *tr_fbuffer (lua_State *L, int pos, float *data) {
  int i;
  lua_Number x;
  /* traverse tr */
  for (i = 0; i < TRSIZE; i++) {
    lua_rawgeti(L, pos, i + 1);
    x = lua_tonumber(L, -1);
    data[i] = (x == 0 && !lua_isnumber(L, -1)) ? STD_TR[i] :
      (float) x;
  }
  lua_pop(L, TRSIZE);
  return data;
}


/* take table at pos as an array and get its dimensions; also, check
 * consistency */
static void get_array_dims (lua_State *L, int pos, int *idim, int *jdim) {
  int i;
  luaL_checktype(L, pos, LUA_TTABLE);
  *idim = lua_objlen(L, pos);
  lua_rawgeti(L, pos, 1); /* first row */
  if (lua_type(L, -1) != LUA_TTABLE)
    luaL_error(L, "bad argument #%d (row 1 is not a table)", pos);
  *jdim = lua_objlen(L, -1);
  lua_pop(L, 1);
  for (i = 1; i < *idim; i++) {
    lua_rawgeti(L, pos, i + 1); /* i-th row */
    if (lua_type(L, -1) != LUA_TTABLE)
      luaL_error(L, "bad argument #%d (row %d is not a table)",
          pos, i + 1);
    if (*jdim != (int) lua_objlen(L, -1))
      luaL_argerror(L, pos, "rows have different sizes");
    lua_pop(L, 1);
  }
}

static float *array_fbuffer (lua_State *L, int pos, int idim, int jdim,
    float *data) {
  int i, j;
  /* traverse table at pos as an array */
  for (i = 0; i < idim; i++) {
    lua_rawgeti(L, pos, i + 1); /* push row */
    for (j = 0; j < jdim; j++) {
      lua_rawgeti(L, -1, j + 1);
      data[i * jdim + j] = pgplot_checkfloat(L, -1, pos,
          TABLE_NONNUM_ERROR);
      lua_pop(L, 1);
    }
    lua_pop(L, 1);
  }
  return data;
}

/* {=================================================================
 *    Functions
 * ==================================================================} */

static int arrow_pgplot (lua_State *L) {
  /* stack contains x1, y1 (tail) and x2, y2 (head) of arrow */
  float x1 = (float) luaL_checknumber(L, 1);
  float y1 = (float) luaL_checknumber(L, 2);
  float x2 = (float) luaL_checknumber(L, 3);
  float y2 = (float) luaL_checknumber(L, 4);
  pgarro_(&x1, &y1, &x2, &y2);
  return 0;
}

static int ask_pgplot (lua_State *L) {
  /* stack contains ask status */
  int flag = lua_toboolean(L, 1);
  pgask_(&flag);
  return 0;
}

static int beginbuffer_pgplot (lua_State *L) {
  (void) L; /* avoid warnings */
  pgbbuf_();
  return 0;
}

static int start_pgplot (lua_State *L) {
  /* stack contains file, nxsub and nysub */
  int dev;
  size_t l;
  const char *file = luaL_optlstring(L, 1, "/NULL", &l);
  int nxsub = luaL_optinteger(L, 2, 1);
  int nysub = luaL_optinteger(L, 3, 1);
  /* initialize if necessary */
  pginit_();
  /* close if already open */
  pgend_();
  dev = pgopen_(file, (int) l);
  if (dev <= 0) luaL_error(L, "error opening device: `%s'", file);
  pgsubp_(&nxsub, &nysub);
  lua_pushinteger(L, dev);
  return 1;
}

static int binhist_pgplot (lua_State *L) {
  /* stack contains one table for X, and other for DATA,
   * where #X == #DATA == NBIN, and center */
  int nbin;
  float *px, *pdata;
  int center;
  fbuffer *fb;
  luaL_checktype(L, 1, LUA_TTABLE);
  luaL_checktype(L, 2, LUA_TTABLE);
  center = lua_toboolean(L, 3);
  nbin = lua_objlen(L, 1);
  if (nbin != (int) lua_objlen(L, 2)) luaL_error(L, "table sizes differ");
  fb = getfbuffer(L, FBUF_INDEX, 2 * nbin);
  /* write X and DATA */
  px = table_fbuffer(L, 1, nbin, fb->data);
  pdata = table_fbuffer(L, 2, nbin, fb->data + nbin);
  pgbin_(&nbin, px, pdata, &center); 
  freefbuffer(fb);
  return 0;
}

static int box_pgplot (lua_State *L) {
  /* stack contains xopt, xtick, nxsub, yopt, ytick, and nysub */
  size_t xl;
  const char *xopt = luaL_checklstring(L, 1, &xl);
  float xtick = (float) luaL_checknumber(L, 2);
  int nxsub = luaL_checkinteger(L, 3);
  size_t yl;
  const char *yopt = luaL_checklstring(L, 4, &yl);
  float ytick = (float) luaL_checknumber(L, 5);
  int nysub = luaL_checkinteger(L, 6);
  pgbox_(xopt, &xtick, &nxsub, yopt, &ytick, &nysub, (int) xl, (int) yl);
  return 0;
}

static int circle_pgplot (lua_State *L) {
  /* stack contains xcent, ycent, and radius */
  float xcent = (float) luaL_checknumber(L, 1);
  float ycent = (float) luaL_checknumber(L, 2);
  float radius = (float) luaL_checknumber(L, 3);
  pgcirc_(&xcent, &ycent, &radius);
  return 0;
}

static int close_pgplot (lua_State *L) {
  (void) L; /* avoid warnings */
  pgclos_();
  return 0;
}

static int labelcont_pgplot (lua_State *L) {
  /* stack contains: 2D array a (as table), level c to be labeled, label,
   * and, optionally, spacing parameters intval and minit,
   * ranges i1, i2, j1, j2, and vector tr.
   * Note that this routine should be called after pgcont, with the same
   * parameters a, c, i1, i2, j1, j2, and tr. */
  int asize, idim, jdim;
  float c;
  size_t ll;
  const char *label;
  int intval, minint;
  int i1, i2, j1, j2;
  float *pa, *ptr;
  fbuffer *fb;
  get_array_dims(L, 1, &idim, &jdim);
  asize = idim * jdim;
  c = luaL_checknumber(L, 2);
  label = luaL_checklstring(L, 3, &ll);
  intval = luaL_optinteger(L, 4, 20);
  minint = luaL_optinteger(L, 5, 10);
  i1 = luaL_optinteger(L, 6, 1);
  i2 = luaL_optinteger(L, 7, idim);
  j1 = luaL_optinteger(L, 8, 1);
  j2 = luaL_optinteger(L, 9, jdim);
  /* set float pointers */
  fb = getfbuffer(L, FBUF_INDEX, asize + TRSIZE);
  /* set a */
  pa = array_fbuffer(L, 1, idim, jdim, fb->data);
  /* set tr */
  if (lua_isnoneornil(L, 10)) ptr = STD_TR;
  else {
    luaL_checktype(L, 10, LUA_TTABLE);
    ptr = tr_fbuffer(L, 10, fb->data + asize);
  }
  /* call */
  pgconl_(pa, &idim, &jdim, &i1, &i2, &j1, &j2, &c, ptr, label,
      &intval, &minint, (int) ll);
  freefbuffer(fb);
  return 0;
}

static int cont_pgplot (lua_State *L) {
  /* stack contains: 2D array a (as table), vector (or number) c, and,
   * optionally, boolean to set line style automatically or not,
   * ranges i1, i2, j1, j2, and vector tr. */
  int asize, idim, jdim, nc, ctype, autols;
  int i1, i2, j1, j2;
  float *pa, *pc, *ptr;
  fbuffer *fb;
  get_array_dims(L, 1, &idim, &jdim);
  asize = idim * jdim;
  ctype = lua_type(L, 2);
  if (ctype != LUA_TNUMBER && ctype != LUA_TTABLE)
    luaL_error(L, "bad argument #%d (table or number expected, got %s)",
        2, lua_typename(L, ctype));
  autols = luaL_opt(L, lua_toboolean, 3, 1); /* true by default */
  i1 = luaL_optinteger(L, 4, 1);
  i2 = luaL_optinteger(L, 5, idim);
  j1 = luaL_optinteger(L, 6, 1);
  j2 = luaL_optinteger(L, 7, jdim);
  /* set float pointers */
  nc = (ctype == LUA_TNUMBER) ? 1 : lua_objlen(L, 2);
  if (nc == 0) luaL_argerror(L, 2, "empty table");
  fb = getfbuffer(L, FBUF_INDEX, asize + nc + TRSIZE);
  /* set a */
  pa = array_fbuffer(L, 1, idim, jdim, fb->data);
  /* set c and nc */
  if (ctype == LUA_TNUMBER) {
    pc = fb->data + asize;
    *pc = (float) lua_tonumber(L, 2);
  }
  else pc = table_fbuffer(L, 2, nc, fb->data + asize);
  if (!autols) nc *= -1;
  /* set tr */
  if (lua_isnoneornil(L, 8)) ptr = STD_TR;
  else {
    luaL_checktype(L, 8, LUA_TTABLE);
    ptr = tr_fbuffer(L, 8, fb->data + asize + nc);
  }
  /* call */
  pgcont_(pa, &idim, &jdim, &i1, &i2, &j1, &j2, pc, &nc, ptr);
  freefbuffer(fb);
  return 0;
}

static int colortable_pgplot (lua_State *L) {
  /* stack contains color table C, where #C == nc and C[i] = {l, r, g, b},
   * contrast (opt), and brightness (opt) */
  float *pl, *pr, *pg, *pb;
  float contra = (float) luaL_optnumber(L, 2, 1.0);
  float bright = (float) luaL_optnumber(L, 3, 0.5);
  int i, nc;
  fbuffer *fb;
  luaL_checktype(L, 1, LUA_TTABLE);
  nc = lua_objlen(L, 1);
  if (nc == 0) luaL_argerror(L, 1, "empty table");
  fb = getfbuffer(L, FBUF_INDEX, 4 * nc);
  pl = fb->data;
  pr = pl + nc;
  pg = pr + nc;
  pb = pg + nc;
  /* traverse color table */
  for (i = 0; i < nc; i++) {
    lua_rawgeti(L, 1, i + 1);
    luaL_checktype(L, -1, LUA_TTABLE);
    lua_rawgeti(L, -1, 1);
    pl[i] = pgplot_checkfloat(L, -1, 1, TABLE_NONNUM_ERROR);
    lua_rawgeti(L, -2, 2);
    pr[i] = pgplot_checkfloat(L, -1, 1, TABLE_NONNUM_ERROR);
    lua_rawgeti(L, -3, 3);
    pg[i] = pgplot_checkfloat(L, -1, 1, TABLE_NONNUM_ERROR);
    lua_rawgeti(L, -4, 4);
    pb[i] = pgplot_checkfloat(L, -1, 1, TABLE_NONNUM_ERROR);
    lua_pop(L, 5);
  }
  pgctab_(pl, pr, pg, pb, &nc, &contra, &bright);
  freefbuffer(fb);
  return 0;
}

static int draw_pgplot (lua_State *L) {
  /* stack contains x and y */
  float x = (float) luaL_checknumber(L, 1);
  float y = (float) luaL_checknumber(L, 2);
  pgdraw_(&x, &y);
  return 0;
}

static int endbuffer_pgplot (lua_State *L) {
  (void) L; /* avoid warnings */
  pgebuf_();
  return 0;
}

static int finish_pgplot (lua_State *L) {
  (void) L; /* avoid warnings */
  pgend_();
  return 0;
}

static int environ_pgplot (lua_State *L) {
  /* stack contains xmin, xmax, ymin, and ymax */
  float xmin = (float) luaL_checknumber(L, 1);
  float xmax = (float) luaL_checknumber(L, 2);
  float ymin = (float) luaL_checknumber(L, 3);
  float ymax = (float) luaL_checknumber(L, 4);
  int just = luaL_optinteger(L, 5, 0);
  int axis = luaL_optinteger(L, 6, 0);
  pgenv_(&xmin, &xmax, &ymin, &ymax, &just, &axis);
  return 0;
}

static int erase_pgplot (lua_State *L) {
  (void) L; /* avoid warnings */
  pgeras_();
  return 0;
}

static int funt_pgplot (lua_State *L) {
  /* stack contains fx(t), fy(t), tmin, tmax, and, optionally, 
   * a flag to start a new plot or not, and #sample points */
  lua_Number tmin, tmax;
  int flag, n;
  lua_Number dt, t;
  float x[MAX_NSAMPLES + 1], y[MAX_NSAMPLES + 1];
  float xmin, xmax, ymin, ymax, adj;
  int i;
  /* read params */
  luaL_checktype(L, 1, LUA_TFUNCTION);
  luaL_checktype(L, 2, LUA_TFUNCTION);
  tmin = luaL_checknumber(L, 3);
  tmax = luaL_checknumber(L, 4);
  flag = lua_toboolean(L, 5);
  n = luaL_optinteger(L, 6, STD_NSAMPLES);
  if (n < 1 || n > MAX_NSAMPLES)
    luaL_error(L, "invalid number of samples: %d", n);
  /* start plotting */
  pgbbuf_();
  /* evaluate functions */
  lua_pushvalue(L, 1); /* fx */
  lua_pushnumber(L, tmin);
  lua_call(L, 1, 1); /* fx(tmin) */
  x[0] = pgplot_checkfloat(L, -1, 1, FUNCTION_NONNUM_ERROR);
  lua_pop(L, 1);
  lua_pushvalue(L, 2); /* fy */
  lua_pushnumber(L, tmin);
  lua_call(L, 1, 1); /* fy(tmin) */
  y[0] = pgplot_checkfloat(L, -1, 2, FUNCTION_NONNUM_ERROR);
  lua_pop(L, 1);
  xmin = xmax = x[0];
  ymin = ymax = y[0];
  t = tmin;
  dt = (tmax - tmin) / n;
  for (i = 1; i <= n; i++) {
    t += dt;
    /* fx(t) */
    lua_pushvalue(L, 1);
    lua_pushnumber(L, t);
    lua_call(L, 1, 1);
    x[i] = pgplot_optfloat(L, -1, x[i-1]);
    lua_pop(L, 1);
    /* fy(t) */
    lua_pushvalue(L, 2);
    lua_pushnumber(L, t);
    lua_call(L, 1, 1);
    y[i] = pgplot_optfloat(L, -1, y[i-1]);
    lua_pop(L, 1);
    /* keep track of min and max */
    if (xmin > x[i]) xmin = x[i];
    if (ymin > y[i]) ymin = y[i];
    if (xmax < x[i]) xmax = x[i];
    if (ymax < y[i]) ymax = y[i];
  }
  /* adjust offset */
  adj = 0.05 * (xmax - xmin);
  if (adj == 0) {
    xmin -= 1.0;
    xmax += 1.0;
  }
  else {
    xmin -= adj;
    xmax += adj;
  }
  adj = 0.05 * (ymax - ymin);
  if (adj == 0) {
    ymin -= 1.0;
    ymax += 1.0;
  }
  else {
    ymin -= adj;
    ymax += adj;
  }
  /* set environment (if necessary) */
  if (!flag) {
    int just = 0; /* not justified */
    int axis = 0; /* box and coord labels */
    pgenv_(&xmin, &xmax, &ymin, &ymax, &just, &axis);
  }
  /* draw curve */
  pgmove_(&x[0], &y[0]);
  for (i = 1; i <= n; i++)
    pgdraw_(&x[i], &y[i]);
  pgebuf_();
  return 0;
}

static int funx_pgplot (lua_State *L) {
  /* stack contains f(x), tmin, tmax, and, optionally, 
   * a flag to start a new plot or not, and #sample points */
  lua_Number tmin, tmax;
  int flag, n;
  lua_Number dt, t;
  float x[MAX_NSAMPLES + 1], y[MAX_NSAMPLES + 1];
  float xmin, xmax, ymin, ymax, adj;
  int i;
  /* read params */
  luaL_checktype(L, 1, LUA_TFUNCTION);
  tmin = luaL_checknumber(L, 2);
  tmax = luaL_checknumber(L, 3);
  flag = lua_toboolean(L, 4);
  n = luaL_optinteger(L, 5, STD_NSAMPLES);
  if (n < 1 || n > MAX_NSAMPLES)
    luaL_error(L, "invalid number of samples: %d", n);
  /* start plotting */
  pgbbuf_();
  /* evaluate functions */
  lua_pushvalue(L, 1); /* f */
  lua_pushnumber(L, tmin);
  lua_call(L, 1, 1); /* f(tmin) */
  y[0] = pgplot_checkfloat(L, -1, 1, FUNCTION_NONNUM_ERROR);
  lua_pop(L, 1);
  xmin = (float) tmin;
  xmax = (float) tmax;
  ymin = ymax = y[0];
  x[0] = xmin;
  t = tmin;
  dt = (tmax - tmin) / n;
  for (i = 1; i <= n; i++) {
    t += dt;
    /* x */
    x[i] = (float) t;
    /* f(x) */
    lua_pushvalue(L, 1);
    lua_pushnumber(L, t);
    lua_call(L, 1, 1);
    y[i] = pgplot_optfloat(L, -1, y[i-1]);
    lua_pop(L, 1);
    /* keep track of min and max */
    if (ymin > y[i]) ymin = y[i];
    if (ymax < y[i]) ymax = y[i];
  }
  /* adjust offset */
  adj = 0.05 * (ymax - ymin);
  if (adj == 0) {
    ymin -= 1.0;
    ymax += 1.0;
  }
  else {
    ymin -= adj;
    ymax += adj;
  }
  /* set environment (if necessary) */
  if (!flag) {
    int just = 0; /* not justified */
    int axis = 0; /* box and coord labels */
    pgenv_(&xmin, &xmax, &ymin, &ymax, &just, &axis);
  }
  /* draw curve */
  pgmove_(&x[0], &y[0]);
  for (i = 1; i <= n; i++)
    pgdraw_(&x[i], &y[i]);
  pgebuf_();
  return 0;
}

static int funy_pgplot (lua_State *L) {
  /* stack contains f(y), tmin, tmax, and, optionally, 
   * a flag to start a new plot or not, and #sample points */
  lua_Number tmin, tmax;
  int flag, n;
  lua_Number dt, t;
  float x[MAX_NSAMPLES + 1], y[MAX_NSAMPLES + 1];
  float xmin, xmax, ymin, ymax, adj;
  int i;
  /* read params */
  luaL_checktype(L, 1, LUA_TFUNCTION);
  tmin = luaL_checknumber(L, 2);
  tmax = luaL_checknumber(L, 3);
  flag = lua_toboolean(L, 4);
  n = luaL_optinteger(L, 5, STD_NSAMPLES);
  if (n < 1 || n > MAX_NSAMPLES)
    luaL_error(L, "invalid number of samples: %d", n);
  /* start plotting */
  pgbbuf_();
  /* evaluate functions */
  lua_pushvalue(L, 1); /* f */
  lua_pushnumber(L, tmin);
  lua_call(L, 1, 1); /* f(tmin) */
  x[0] = pgplot_checkfloat(L, -1, 1, FUNCTION_NONNUM_ERROR);
  lua_pop(L, 1);
  ymin = (float) tmin;
  ymax = (float) tmax;
  xmin = xmax = x[0];
  y[0] = ymin;
  t = tmin;
  dt = (tmax - tmin) / n;
  for (i = 1; i <= n; i++) {
    t += dt;
    /* y */
    y[i] = (float) t;
    /* f(y) */
    lua_pushvalue(L, 1);
    lua_pushnumber(L, t);
    lua_call(L, 1, 1);
    x[i] = pgplot_optfloat(L, -1, x[i-1]);
    lua_pop(L, 1);
    /* keep track of min and max */
    if (xmin > x[i]) xmin = x[i];
    if (xmax < x[i]) xmax = x[i];
  }
  /* adjust offset */
  adj = 0.05 * (xmax - xmin);
  if (adj == 0) {
    xmin -= 1.0;
    xmax += 1.0;
  }
  else {
    xmin -= adj;
    xmax += adj;
  }
  /* set environment (if necessary) */
  if (!flag) {
    int just = 0; /* not justified */
    int axis = 0; /* box and coord labels */
    pgenv_(&xmin, &xmax, &ymin, &ymax, &just, &axis);
  }
  /* draw curve */
  pgmove_(&x[0], &y[0]);
  for (i = 1; i <= n; i++)
    pgdraw_(&x[i], &y[i]);
  pgebuf_();
  return 0;
}

static int grayimage_pgplot (lua_State *L) {
  /* stack contains: 2D array a (as table), and, optionally, 
   * BG and FG levels, ranges i1, i2, j1, j2, and vector tr. */
  int asize, idim, jdim, i1, i2, j1, j2;
  float *pa, *ptr;
  float bg, fg;
  fbuffer *fb;
  get_array_dims(L, 1, &idim, &jdim);
  asize = idim * jdim;
  /* set a */
  fb = getfbuffer(L, FBUF_INDEX, asize + TRSIZE);
  pa = array_fbuffer(L, 1, idim, jdim, fb->data);
  /* set bg and fg */
  if (lua_isnoneornil(L, 2)) { /* set bg as min(a) */
    int i;
    bg = pa[0];
    for (i = 1; i < asize; i++)
      if (bg > pa[i]) bg = pa[i];
  }
  else bg = (float) luaL_checknumber(L, 2);
  if (lua_isnoneornil(L, 3)) { /* set fg as max(a) */
    int i;
    fg = pa[0];
    for (i = 1; i < asize; i++)
      if (fg < pa[i]) fg = pa[i];
  }
  else fg = (float) luaL_checknumber(L, 3);
  i1 = luaL_optinteger(L, 4, 1);
  i2 = luaL_optinteger(L, 5, idim);
  j1 = luaL_optinteger(L, 6, 1);
  j2 = luaL_optinteger(L, 7, jdim);
  /* set tr */
  if (lua_isnoneornil(L, 8)) ptr = STD_TR;
  else {
    luaL_checktype(L, 8, LUA_TTABLE);
    ptr = tr_fbuffer(L, 8, fb->data + asize);
  }
  /* call */
  pggray_(pa, &idim, &jdim, &i1, &i2, &j1, &j2, &fg, &bg, ptr);
  freefbuffer(fb);
  return 0;
}

static int hist_pgplot (lua_State *L) {
  /* stack contains data (as table), nbin, and optionally
   * flag to start new env or not, datmin, and datmax */
  float datmin, datmax, *pdata;
  int n, nbin, flag;
  fbuffer *fb;
  luaL_checktype(L, 1, LUA_TTABLE);
  n = lua_objlen(L, 1);
  if (n == 0) luaL_argerror(L, 1, "empty table");
  nbin = luaL_checkinteger(L, 2);
  flag = lua_toboolean(L, 3);
  fb = getfbuffer(L, FBUF_INDEX, n);
  pdata = table_fbuffer(L, 1, n, fb->data);
  /* compute datmin */
  if (lua_isnoneornil(L, 4)) {
    int i;
    datmin = pdata[0];
    for (i = 1; i < n; i++)
      if (datmin > pdata[i]) datmin = pdata[i];
  }
  else datmin = (float) luaL_checknumber(L, 4);
  /* compute datmax */
  if (lua_isnoneornil(L, 5)) {
    int i;
    datmax = pdata[0];
    for (i = 1; i < n; i++)
      if (datmax < pdata[i]) datmax = pdata[i];
  }
  else datmax = (float) luaL_checknumber(L, 5);
  /* call */
  pghist_(&n, pdata, &datmin, &datmax, &nbin, &flag);
  freefbuffer(fb);
  return 0;
}

static int identify_pgplot (lua_State *L) {
  (void) L; /* avoid warnings */
  pgiden_();
  return 0;
}

static int colorimage_pgplot (lua_State *L) {
  /* stack contains: 2D array a (as table), and, optionally, 
   * low and high levels, ranges i1, i2, j1, j2, and vector tr. */
  int asize, idim, jdim, i1, i2, j1, j2;
  float *pa, *ptr;
  float a1, a2;
  fbuffer *fb;
  get_array_dims(L, 1, &idim, &jdim);
  asize = idim * jdim;
  /* set a */
  fb = getfbuffer(L, FBUF_INDEX, asize + TRSIZE);
  pa = array_fbuffer(L, 1, idim, jdim, fb->data);
  /* set a1 and a2 */
  if (lua_isnoneornil(L, 2)) { /* set a1 as min(a) */
    int i;
    a1 = pa[0];
    for (i = 1; i < asize; i++)
      if (a1 > pa[i]) a1 = pa[i];
  }
  else a1 = (float) luaL_checknumber(L, 2);
  if (lua_isnoneornil(L, 3)) { /* set a2 as max(a) */
    int i;
    a2 = pa[0];
    for (i = 1; i < asize; i++)
      if (a2 < pa[i]) a2 = pa[i];
  }
  else a2 = (float) luaL_checknumber(L, 3);
  i1 = luaL_optinteger(L, 4, 1);
  i2 = luaL_optinteger(L, 5, idim);
  j1 = luaL_optinteger(L, 6, 1);
  j2 = luaL_optinteger(L, 7, jdim);
  /* set tr */
  if (lua_isnoneornil(L, 8)) ptr = STD_TR;
  else {
    luaL_checktype(L, 8, LUA_TTABLE);
    ptr = tr_fbuffer(L, 8, fb->data + asize);
  }
  /* call */
  pgimag_(pa, &idim, &jdim, &i1, &i2, &j1, &j2, &a1, &a2, ptr);
  freefbuffer(fb);
  return 0;
}

static int label_pgplot (lua_State *L) {
  /* stack contains xlabel, ylabel and toplabel */
  const char *xlbl, *ylbl, *tlbl;
  size_t xl, yl, tl;
  xlbl = luaL_checklstring(L, 1, &xl);
  ylbl = luaL_checklstring(L, 2, &yl);
  tlbl = luaL_checklstring(L, 3, &tl);
  pglab_(xlbl, ylbl, tlbl, (int) xl, (int) yl, (int) tl);
  return 0;
}

static int strlen_pgplot (lua_State *L) {
  /* stack contains string and optionally units */
  size_t sl;
  const char *string = luaL_checklstring(L, 1, &sl);
  int units = luaL_optinteger(L, 2, 0); /* default: norm. dev. coords */
  float xl, yl;
  pglen_(&units, string, &xl, &yl, (int) sl);
  lua_pushnumber(L, (lua_Number) xl);
  lua_pushnumber(L, (lua_Number) yl);
  return 2;
}

static int line_pgplot (lua_State *L) {
  /* stack contains xpts and ypts */
  float *px, *py;
  int n;
  fbuffer *fb;
  luaL_checktype(L, 1, LUA_TTABLE);
  luaL_checktype(L, 2, LUA_TTABLE);
  n = lua_objlen(L, 1);
  if (n != (int) lua_objlen(L, 2)) luaL_error(L, "table sizes differ");
  /* set floats */
  fb = getfbuffer(L, FBUF_INDEX, 2 * n);
  px = table_fbuffer(L, 1, n, fb->data);
  py = table_fbuffer(L, 2, n, fb->data + n);
  /* call */
  pgline_(&n, px, py);
  freefbuffer(fb);
  return 0;
}

static int move_pgplot (lua_State *L) {
  float x = (float) luaL_checknumber(L, 1);
  float y = (float) luaL_checknumber(L, 2);
  pgmove_(&x, &y);
  return 0;
}

static int mtext_pgplot (lua_State *L) {
  /* stack contains text, and optionally, side, disp, coord, and fjust */
  size_t tl;
  const char *text = luaL_checklstring(L, 1, &tl);
  size_t sl;
  const char *side = luaL_checklstring(L, 2, &sl);
  float disp = (float) luaL_optnumber(L, 3, 1.0);
  float coord = (float) luaL_optnumber(L, 4, 0.0);
  float fjust = (float) luaL_optnumber(L, 5, 0.0);
  pgmtxt_(side, &disp, &coord, &fjust, text, (int) sl, (int) tl);
  return 0;
}

static int open_pgplot (lua_State *L) {
  size_t dl;
  const char *device = luaL_checklstring(L, 1, &dl);
  int dev = pgopen_(device, (int) dl);
  lua_pushinteger(L, dev);
  return 1;
}

static int newpage_pgplot (lua_State *L) {
  (void) L; /* avoid warnings */
  pgpage_();
  return 0;
}

static int panel_pgplot (lua_State *L) {
  int ix = luaL_optinteger(L, 1, 1);
  int iy = luaL_optinteger(L, 2, 1);
  pgpanl_(&ix, &iy);
  return 0;
}

static int setpapersize_pgplot (lua_State *L) {
  float width = (float) luaL_checknumber(L, 1);
  float height = (float) luaL_checknumber(L, 2);
  float ar = height / width;
  pgpap_(&width, &ar);
  return 0;
}

static int polygon_pgplot (lua_State *L) {
  /* stack contains xpts, ypts */
  float *px, *py;
  int n;
  fbuffer *fb;
  luaL_checktype(L, 1, LUA_TTABLE);
  luaL_checktype(L, 2, LUA_TTABLE);
  n = lua_objlen(L, 1);
  if (n != (int) lua_objlen(L, 2)) luaL_error(L, "table sizes differ");
  /* set floats */
  fb = getfbuffer(L, FBUF_INDEX, 2 * n);
  px = table_fbuffer(L, 1, n, fb->data);
  py = table_fbuffer(L, 2, n, fb->data + n);
  /* call */
  pgpoly_(&n, px, py);
  freefbuffer(fb);
  return 0;
}

static int point_pgplot (lua_State *L) {
  /* stack contains xpts, ypts, and symbol */
  float *px, *py;
  int n, symbol;
  fbuffer *fb;
  luaL_checktype(L, 1, LUA_TTABLE);
  luaL_checktype(L, 2, LUA_TTABLE);
  n = lua_objlen(L, 1);
  if (n != (int) lua_objlen(L, 2)) luaL_error(L, "table sizes differ");
  symbol = luaL_optinteger(L, 3, STD_SYMBOL);
  /* set floats */
  fb = getfbuffer(L, FBUF_INDEX, 2 * n); 
  px = table_fbuffer(L, 1, n, fb->data);
  py = table_fbuffer(L, 2, n, fb->data + n);
  /* call */
  pgpt_(&n, px, py, &symbol);
  freefbuffer(fb);
  return 0;
}

static int ptext_pgplot (lua_State *L) {
  /* stack contains text, x, y, and optionally angle and fjust */
  size_t tl;
  const char *text = luaL_checklstring(L, 1, &tl);
  float x = (float) luaL_checknumber(L, 2);
  float y = (float) luaL_checknumber(L, 3);
  float angle = (float) luaL_optnumber(L, 4, 0.0);
  float fjust = (float) luaL_optnumber(L, 5, 0.0);
  pgptxt_(&x, &y, &angle, &fjust, text, (int) tl);
  return 0;
}

static int getarrowstyle_pgplot (lua_State *L) {
  float angle, barb;
  int fs;
  pgqah_(&fs, &angle, &barb);
  if (fs == 2) fs = 0; /* adjust to boolean */
  lua_pushnumber(L, (lua_Number) angle);
  lua_pushnumber(L, (lua_Number) barb);
  lua_pushboolean(L, fs);
  return 3;
}

static int getcharfont_pgplot (lua_State *L) {
  int font;
  pgqcf_(&font);
  lua_pushinteger(L, font);
  return 1;
}

static int getcharheight_pgplot (lua_State *L) {
  float size;
  pgqch_(&size);
  lua_pushnumber(L, (lua_Number) size);
  return 1;
}

static int getcolorindex_pgplot (lua_State *L) {
  int ci;
  pgqci_(&ci);
  lua_pushinteger(L, ci);
  return 1;
}

static int getcolorindexrange_pgplot (lua_State *L) {
  int icilo, icihi;
  pgqcir_(&icilo, &icihi);
  lua_pushinteger(L, icilo);
  lua_pushinteger(L, icihi);
  return 2;
}

static int getclippingstatus_pgplot (lua_State *L) {
  int state;
  pgqclp_(&state);
  lua_pushboolean(L, state);
  return 1;
}

static int getcolorcap_pgplot (lua_State *L) {
  int ci1, ci2;
  pgqcol_(&ci1, &ci2);
  lua_pushinteger(L, ci1);
  lua_pushinteger(L, ci2);
  return 2;
}

static int getcolorrep_pgplot (lua_State *L) {
  int ci = luaL_checkinteger(L, 1);
  float cr, cg, cb;
  pgqcr_(&ci, &cr, &cg, &cb);
  lua_pushnumber(L, (lua_Number) cr);
  lua_pushnumber(L, (lua_Number) cg);
  lua_pushnumber(L, (lua_Number) cb);
  return 3;
}

static int getfillstyle_pgplot (lua_State *L) {
  int fs;
  pgqfs_(&fs);
  lua_pushinteger(L, fs);
  return 1;
}

static int gethatchingstyle_pgplot (lua_State *L) {
  float angle, sepn, phase;
  pgqhs_(&angle, &sepn, &phase);
  lua_pushnumber(L, (lua_Number) angle);
  lua_pushnumber(L, (lua_Number) sepn);
  lua_pushnumber(L, (lua_Number) phase);
  return 3;
}

static int getcurdevice_pgplot (lua_State *L) {
  int id;
  pgqid_(&id);
  lua_pushinteger(L, id);
  return 1;
}

static int getITF_pgplot (lua_State *L) {
  int itf; /* image transfer function */
  pgqitf_(&itf);
  lua_pushinteger(L, itf);
  return 1;
}

static int getlinestyle_pgplot (lua_State *L) {
  int ls;
  pgqls_(&ls);
  lua_pushinteger(L, ls);
  return 1;
}

static int getlinewidth_pgplot (lua_State *L) {
  int lw;
  pgqlw_(&lw);
  lua_pushinteger(L, lw);
  return 1;
}

static int getpenposition_pgplot (lua_State *L) {
  float x, y;
  pgqpos_(&x, &y);
  lua_pushnumber(L, (lua_Number) x);
  lua_pushnumber(L, (lua_Number) y);
  return 2;
}

static int gettextbg_pgplot (lua_State *L) {
  int tbci;
  pgqtbg_(&tbci);
  lua_pushinteger(L, tbci);
  return 1;
}

static int getviewport_pgplot (lua_State *L) {
  float x1, x2, y1, y2;
  int units = luaL_optinteger(L, 1, 0); /* def.: norm. dev. coords */
  pgqvp_(&units, &x1, &x2, &y1, &y2);
  lua_pushnumber(L, (lua_Number) x1);
  lua_pushnumber(L, (lua_Number) x2);
  lua_pushnumber(L, (lua_Number) y1);
  lua_pushnumber(L, (lua_Number) y2);
  return 4;
}

static int getviewsize_pgplot (lua_State *L) {
  float x1, x2, y1, y2;
  int units = luaL_optinteger(L, 1, 0); /* def.: norm. dev. coords */
  pgqvsz_(&units, &x1, &x2, &y1, &y2);
  lua_pushnumber(L, (lua_Number) x1);
  lua_pushnumber(L, (lua_Number) x2);
  lua_pushnumber(L, (lua_Number) y1);
  lua_pushnumber(L, (lua_Number) y2);
  return 4;
}

static int getwindow_pgplot (lua_State *L) {
  float x1, x2, y1, y2;
  pgqwin_(&x1, &x2, &y1, &y2);
  lua_pushnumber(L, (lua_Number) x1);
  lua_pushnumber(L, (lua_Number) x2);
  lua_pushnumber(L, (lua_Number) y1);
  lua_pushnumber(L, (lua_Number) y2);
  return 4;
}

static int rectangle_pgplot (lua_State *L) {
  /* stack contains x1, x2, y1, y2 */
  float x1 = (float) luaL_checknumber(L, 1);
  float x2 = (float) luaL_checknumber(L, 2);
  float y1 = (float) luaL_checknumber(L, 3);
  float y2 = (float) luaL_checknumber(L, 4);
  pgrect_(&x1, &x2, &y1, &y2);
  return 0;
}

static int save_pgplot (lua_State *L) {
  (void) L; /* avoid warnings */
  pgsave_();
  return 0;
}

static int unsave_pgplot (lua_State *L) {
  (void) L; /* avoid warnings */
  pgunsa_();
  return 0;
}

static int setarrowstyle_pgplot (lua_State *L) {
  /* stack: angle, barb, and (opt) fs (boolean). */
  float angle = (float) luaL_checknumber(L, 1);
  float barb = (float) luaL_checknumber(L, 2);
  int fs = (lua_isnoneornil(L, 3)) ? 1 : lua_toboolean(L, 3);
  if (!fs) fs = 2; /* not filled */ 
  pgsah_(&fs, &angle, &barb);
  return 0;
}

static int setcharfont_pgplot (lua_State *L) {
  int font = luaL_optinteger(L, 1, 1); /* default: normal font */
  pgscf_(&font);
  return 0;
}

static int setcharheight_pgplot (lua_State *L) {
  float size = (float) luaL_optnumber(L, 1, 1.0);
  pgsch_(&size);
  return 0;
}

static int setcolorindex_pgplot (lua_State *L) {
  int ci = luaL_optinteger(L, 1, 1);
  pgsci_(&ci);
  return 0;
}

static int setcolorindexrange_pgplot (lua_State *L) {
  int icilo = luaL_checkinteger(L, 1);
  int icihi = luaL_checkinteger(L, 2);
  pgscir_(&icilo, &icihi);
  return 0;
}

static int setclippingstatus_pgplot (lua_State *L) {
  int state = lua_toboolean(L, 1);
  pgsclp_(&state);
  return 0;
}

static int setcolorrep_pgplot (lua_State *L) {
  /* stack contains ci, cr, cg, and cb */
  int ci = luaL_checkinteger(L, 1);
  float cr = (float) luaL_checknumber(L, 2);
  float cg = (float) luaL_checknumber(L, 3);
  float cb = (float) luaL_checknumber(L, 4);
  pgscr_(&ci, &cr, &cg, &cb);
  return 0;
}

static int scroll_pgplot (lua_State *L) {
  float dx = (float) luaL_optnumber(L, 1, 0.0);
  float dy = (float) luaL_optnumber(L, 2, 0.0);
  pgscrl_(&dx, &dy);
  return 0;
}

static int setcolorbyname_pgplot (lua_State *L) {
  /* stack contains ci and name */
  int ci = luaL_checkinteger(L, 1);
  size_t nl;
  const char *name = luaL_checklstring(L, 2, &nl);
  int ier;
  pgscrn_(&ci, name, &ier, (int) nl);
  if (ier)
    luaL_error(L, "file (rgb.txt) or color (`%s') not found", name);
  return 0;
}

static int setfillstyle_pgplot (lua_State *L) {
  int fs = luaL_optinteger(L, 1, 1);
  pgsfs_(&fs);
  return 0;
}

static int sethatchingstyle_pgplot (lua_State *L) {
  /* stack contains angle, sepn, and phase */
  float angle = (float) luaL_checknumber(L, 1);
  float sepn = (float) luaL_checknumber(L, 2);
  float phase = (float) luaL_checknumber(L, 3);
  pgshs_(&angle, &sepn, &phase);
  return 0;
}

static int setITF_pgplot (lua_State *L) {
  /* image transfer function:
   * itf = 0: linear
   *       1: logarithm
   *       2: square root */
  int itf = luaL_optinteger(L, 1, 0);
  pgsitf_(&itf);
  return 0;
}

static int select_pgplot (lua_State *L) {
  int id = luaL_checkinteger(L, 1);
  pgslct_(&id);
  return 0;
}

static int setlinestyle_pgplot (lua_State *L) {
  int ls = luaL_optinteger(L, 1, 1);
  pgsls_(&ls);
  return 0;
}

static int setlinewidth_pgplot (lua_State *L) {
  int lw = luaL_optinteger(L, 1, 1);
  pgslw_(&lw);
  return 0;
}

static int settextbg_pgplot (lua_State *L) {
  int tbci = luaL_optinteger(L, 1, -1); /* transparent by default */
  pgstbg_(&tbci);
  return 0;
}

static int subdivide_pgplot (lua_State *L) {
  int nxsub = luaL_optinteger(L, 1, 1);
  int nysub = luaL_optinteger(L, 2, 1);
  pgsubp_(&nxsub, &nysub);
  return 0;
}

static int setviewport_pgplot (lua_State *L) {
  /* stack contains xleft, xright, ybot, and ytop */
  float xleft = (float) luaL_checknumber(L, 1);
  float xright = (float) luaL_checknumber(L, 2);
  float ybot = (float) luaL_checknumber(L, 3);
  float ytop = (float) luaL_checknumber(L, 4);
  pgsvp_(&xleft, &xright, &ybot, &ytop);
  return 0;
}

static int setwindow_pgplot (lua_State *L) {
  /* stack contains x1, x2, y1, and y2 */
  float x1 = (float) luaL_checknumber(L, 1);
  float x2 = (float) luaL_checknumber(L, 2);
  float y1 = (float) luaL_checknumber(L, 3);
  float y2 = (float) luaL_checknumber(L, 4);
  pgswin_(&x1, &x2, &y1, &y2);
  return 0;
}

static int text_pgplot (lua_State *L) {
  /* stack contains text, x, and y */
  size_t tl;
  const char *text = luaL_checklstring(L, 1, &tl);
  float x = (float) luaL_checknumber(L, 2);
  float y = (float) luaL_checknumber(L, 3);
  pgtext_(&x, &y, text, (int) tl);
  return 0;
}

static int update_pgplot (lua_State *L) {
  (void) L; /* avoid warnings */
  pgupdt_();
  return 0;
}

static int setviewsize_pgplot (lua_State *L) {
  float xleft = (float) luaL_checknumber(L, 1);
  float xright = (float) luaL_checknumber(L, 2);
  float ybot = (float) luaL_checknumber(L, 3);
  float ytop = (float) luaL_checknumber(L, 4);
  pgvsiz_(&xleft, &xright, &ybot, &ytop);
  return 0;
}

static int setwindowvp_pgplot (lua_State *L) {
  /* stack contains x1, x2, y1, y2 */
  float x1 = (float) luaL_checknumber(L, 1);
  float x2 = (float) luaL_checknumber(L, 2);
  float y1 = (float) luaL_checknumber(L, 3);
  float y2 = (float) luaL_checknumber(L, 4);
  pgwnad_(&x1, &x2, &y1, &y2);
  return 0;
}


/* {=================================================================
 *    Interface
 * ==================================================================} */

static const luaL_reg pgplot_lib[] = {
  {"arrow", arrow_pgplot},
  {"ask", ask_pgplot},
  {"beginbuffer", beginbuffer_pgplot},
  {"start", start_pgplot},
  {"binhist", binhist_pgplot},
  {"box", box_pgplot},
  {"circle", circle_pgplot},
  {"close", close_pgplot},
  {"labelcont", labelcont_pgplot},
  {"cont", cont_pgplot},
  {"colortable", colortable_pgplot},
  {"draw", draw_pgplot},
  {"endbuffer", endbuffer_pgplot},
  {"finish", finish_pgplot},
  {"environ", environ_pgplot},
  {"erase", erase_pgplot},
  {"funt", funt_pgplot},
  {"funx", funx_pgplot},
  {"funy", funy_pgplot},
  {"grayimage", grayimage_pgplot},
  {"hist", hist_pgplot},
  {"identify", identify_pgplot},
  {"colorimage", colorimage_pgplot},
  {"label", label_pgplot},
  {"strlen", strlen_pgplot},
  {"line", line_pgplot},
  {"move", move_pgplot},
  {"mtext", mtext_pgplot},
  {"open", open_pgplot},
  {"newpage", newpage_pgplot},
  {"panel", panel_pgplot},
  {"setpapersize", setpapersize_pgplot},
  {"polygon", polygon_pgplot},
  {"point", point_pgplot},
  {"ptext", ptext_pgplot},
  {"getarrowstyle", getarrowstyle_pgplot},
  {"getcharfont", getcharfont_pgplot},
  {"getcharheight", getcharheight_pgplot},
  {"getcolorindex", getcolorindex_pgplot},
  {"getcolorindexrange", getcolorindexrange_pgplot},
  {"getclippingstatus", getclippingstatus_pgplot},
  {"getcolorcap", getcolorcap_pgplot},
  {"getcolorrep", getcolorrep_pgplot},
  {"getfillstyle", getfillstyle_pgplot},
  {"gethatchingstyle", gethatchingstyle_pgplot},
  {"getcurdevice", getcurdevice_pgplot},
  {"getITF", getITF_pgplot},
  {"getlinestyle", getlinestyle_pgplot},
  {"getlinewidth", getlinewidth_pgplot},
  {"getpenposition", getpenposition_pgplot},
  {"gettextbg", gettextbg_pgplot},
  {"getviewport", getviewport_pgplot},
  {"getviewsize", getviewsize_pgplot},
  {"getwindow", getwindow_pgplot},
  {"rectangle", rectangle_pgplot},
  {"save", save_pgplot},
  {"unsave", unsave_pgplot},
  {"setarrowstyle", setarrowstyle_pgplot},
  {"setcharfont", setcharfont_pgplot},
  {"setcharheight", setcharheight_pgplot},
  {"setcolorindex", setcolorindex_pgplot},
  {"setcolorindexrange", setcolorindexrange_pgplot},
  {"setclippingstatus", setclippingstatus_pgplot},
  {"setcolorrep", setcolorrep_pgplot},
  {"scroll", scroll_pgplot},
  {"setcolorbyname", setcolorbyname_pgplot},
  {"setfillstyle", setfillstyle_pgplot},
  {"sethatchingstyle", sethatchingstyle_pgplot},
  {"setITF", setITF_pgplot},
  {"select", select_pgplot},
  {"setlinestyle", setlinestyle_pgplot},
  {"setlinewidth", setlinewidth_pgplot},
  {"settextbg", settextbg_pgplot},
  {"subdivide", subdivide_pgplot},
  {"setviewport", setviewport_pgplot},
  {"setwindow", setwindow_pgplot},
  {"text", text_pgplot},
  {"update", update_pgplot},
  {"setviewsize", setviewsize_pgplot},
  {"setwindowvp", setwindowvp_pgplot},
  {NULL, NULL}
};

static int pgplotMT__gc (lua_State *L) {
  finish_pgplot(L);
  return 0;
}

int luaopen_luapgplot (lua_State *L) {
  lua_newtable(L); /* environment */
  lua_replace(L, LUA_ENVIRONINDEX);
  lua_createtable(L, 0, 80 + 2);
  luaL_register(L, LUAPGPLOT_LIBNAME, pgplot_lib);
  lua_pushstring(L, LUAPGPLOT_VERSION);
  lua_setfield(L, -2, "_VERSION");
  pgplot_avdevices(L);
  lua_setfield(L, -2, "AV_DEVICES"); /* available devices */
  /* class MT */
  lua_createtable(L, 0, 2);
  lua_pushcfunction(L, pgplotMT__gc);
  lua_setfield(L, -2, "__gc");
  lua_pushvalue(L, -2);
  lua_setfield(L, -2, "__metatable");
  lua_setmetatable(L, -2);
  return 1;
}

