-- {=====================================================================
--
-- Plotting library for Lua, based on PGPLOT
-- See Copyright Notice in luapgplot.h
--
-- $Id: $
--
-- =====================================================================}

-- cache locally what is needed
local assert = assert
local error = error
local ipairs = ipairs
local tonumber = tonumber
local tostring = tostring
local type = type
local format = string.format
local getn = table.getn
local mod = math.mod
local PI = math.pi
local sin = math.sin
local cos = math.cos

local pgplot = require 'luapgplot'
module 'pgplot'

-- auxiliar functions
local check_arg = function(x, narg, t, opt)
	if x then
		if t == 'string' then x = tostring(x) end
		local tx = type(x)
		assert(tx == t,
			format('bad argument #%d: %s expected, got %s',
					narg, t, tx))
		return x
	else
		if t == 'string' and not opt then opt = '' end
		return opt
	end
end

local check_device = function()
	assert(pgplot.getcurdevice() > 0, 'device not initialized')
end

-- shortcut to initialize /XWINDOW
startx = function()
	local xdev
	for _, dev in ipairs(pgplot.AV_DEVICES) do
		if dev.type == '/GW' or dev.type == '/XWINDOW' then
			xdev = dev.type
			break
		end
	end
	assert(xdev, 'can\'t initialize device')
	pgplot.start(xdev)
	pgplot.setcolorbyname(0, 'white')
	pgplot.setcolorbyname(1, 'black')
end

-- constants
SAMEPLOT = false

-- =====================================================================
-- Image related

-- =======   Palettes   =======
-- default # palette levels
local STD_LEVELS = 255
-- generate palette from color (r1, g1, b1) to (r2, g2, b2) with maxl levels
genpalette = function(r1, g1, b1, r2, g2, b2, maxl)
	local pal = {}
	local maxl = maxl or STD_LEVELS
	local dl = 1 / maxl
	local dr = (r2 - r1) / maxl
	local dg = (g2 - g1) / maxl
	local db = (b2 - b1) / maxl
	for i = 0, maxl do
		pal[i + 1] = {i * dl, r1 + i * dr, g1 + i * dg, b1 + i * db}
	end
	return pal
end
-- default palette
PALETTE = genpalette(0, 0, 0, 1, 1, 1) -- gray levels

-- =======   Image   =======
image = function(data, pal)
	check_device()
	local data = check_arg(data, 1, 'table')
	local pal = check_arg(pal, 2, 'table', PALETTE)
	local nx = getn(data)
	assert(nx > 0, 'bad argument #1: empty table')
	local ny = getn(data[1])
	assert(ny > 0, 'bad argument #1: empty row in table')
	for i = 2, nx do
		assert(ny == getn(data[i]), 'table sizes differ')
	end
	pgplot.environ(0.5, nx + 0.5, 0.5, ny + 0.5, 0, -1) -- only box
	pgplot.colortable(pal)
	pgplot.colorimage(data)
end

-- =======   Contour   =======
contour = function(data, levels, labels, sameplot)
	check_device()
	local sameplot = sameplot == nil and SAMEPLOT or sameplot
	local data = check_arg(data, 1, 'table')
	local nx = getn(data)
	assert(nx > 0, 'bad argument #1: empty table')
	local ny = getn(data[1])
	assert(ny > 0, 'bad argument #1: empty row in table')
	for i = 2, nx do
		assert(ny == getn(data[i]), 'table sizes differ')
	end
	local tlevels = type(levels)
	if tlevels == 'number' then
		if not sameplot then
			pgplot.environ(0.5, nx + 0.5, 0.5, ny + 0.5, 0, -1)
		end
		pgplot.cont(data, levels)
		if labels then
			local label = check_arg(labels, 3, 'string',
					format('%.2f', tostring(levels)))
			pgplot.labelcont(data, levels, label)
		end
	elseif tlevels == 'table' then
		local nl = getn(levels)
		assert(nl > 0, 'bad argument #2: empty table')
		local tlabels = type(labels)
		if tlabels == 'table' then
			assert(getn(labels) == nl, 'level and label sizes differ')
		end
		if not sameplot then
			pgplot.environ(0.5, nx + 0.5, 0.5, ny + 0.5, 0, -1)
		end
		-- draw levels and labels
		pgplot.beginbuffer()
		for i = 1, nl do
			local level = tonumber(levels[i])
			if not level then
				pgplot.endbuffer()
				error('level is not numeric')
			end
			pgplot.cont(data, tonumber(levels[i]))
			if labels then
				local label = tlabels == 'table' and
						labels[i] or format('%.2f', tostring(levels[i]))
				pgplot.labelcont(data, levels[i], label)
			end
		end
		pgplot.endbuffer()
	else
		error('bad argument #2: table or number expected, got ' .. tlevels)
	end
end


-- =====================================================================
-- Function related

local TMIN = -10
local TMAX = 10
local STD_NSAMPLES = 100
local STD_OFFSET = 0.05

-- auxiliar to find adjusted min,max_{tmin <= t <= tmax}(f(t))
local find_fminmax = function(f, tmin, tmax, nsamples)
	assert(nsamples > 0, '#samples is non-positive')
	local t = tmin
	local dt = (tmax - tmin) / nsamples
	local ft = f(t)
	local fmin = ft
	local fmax = ft
	for i = 1, nsamples do
		t = t + dt
		ft = f(t)
		if fmin > ft then fmin = ft end
		if fmax < ft then fmax = ft end
	end
	local adj = STD_OFFSET * (fmax - fmin)
	if (adj == 0) then
		fmin = fmin - 1.0
		fmax = fmax + 1.0
	else
		fmin = fmin - adj
		fmax = fmax + adj
	end
	return fmin, fmax
end

-- plot function
func = function(f, xmin, xmax, nsamples, xlbl, ylbl, title)
	check_device()
	local xmin = check_arg(xmin, 2, 'number', TMIN)
	local xmax = check_arg(xmax, 3, 'number', TMAX)
	local nsamples = check_arg(nsamples, 4, 'number', STD_NSAMPLES)
	local xlbl = check_arg(xlbl, 5, 'string')
	local ylbl = check_arg(ylbl, 6, 'string')
	local title = check_arg(title, 7, 'string')
	-- set colors
	local _, mc = pgplot.getcolorcap()
	if type(f) == 'function' then
		local fmin, fmax = find_fminmax(f, xmin, xmax, nsamples)
		pgplot.environ(xmin, xmax, fmin, fmax)
		pgplot.label(xlbl, ylbl, title)
		if mc > 1 then pgplot.setcolorindex(2) end
		pgplot.funx(f, xmin, xmax, true, nsamples)
		pgplot.setcolorindex()
	elseif type(f) == 'table' then
		local n = getn(f)
		assert(n > 0, 'bad argument #1: empty table')
		-- get min and max for all functions
		assert(type(f[1]) == 'function', 'table value is not a function')
		local fmin, fmax = find_fminmax(f[1], xmin, xmax, nsamples)
		for i = 2, n do
			assert(type(f[i]) == 'function', 'table value is not a function')
			local lmin, lmax = find_fminmax(f[i], xmin, xmax, nsamples)
			if fmin > lmin then fmin = lmin end
			if fmax < lmax then fmax = lmax end
		end
		-- plot
		pgplot.environ(xmin, xmax, fmin, fmax)
		pgplot.label(xlbl, ylbl, title)
		pgplot.beginbuffer()
		for i = 1, n do
			if mc > 1 then
				pgplot.setcolorindex(mod(i - 1, mc - 1) + 2)
			end
			pgplot.funx(f[i], xmin, xmax, true, nsamples)
		end
		pgplot.setcolorindex()
		pgplot.endbuffer()
	else
		error('bad argument #1: table or function expected, got ' .. type(f))
	end
end

parametric = function(fx, fy, tmin, tmax, nsamples, xlbl, ylbl, title)
	check_device()
	local fx = check_arg(fx, 1, 'function')
	local fy = check_arg(fy, 2, 'function')
	local tmin = check_arg(tmin, 3, 'number', TMIN)
	local tmax = check_arg(tmax, 4, 'number', TMAX)
	local nsamples = check_arg(nsamples, 5, 'number', STD_NSAMPLES)
	local xlbl = check_arg(xlbl, 6, 'string')
	local ylbl = check_arg(ylbl, 7, 'string')
	local title = check_arg(title, 8, 'string')
	local _, mc = pgplot.getcolorcap()
	-- plot
	local fxmin, fxmax = find_fminmax(fx, tmin, tmax, nsamples)
	local fymin, fymax = find_fminmax(fy, tmin, tmax, nsamples)
	pgplot.environ(fxmin, fxmax, fymin, fymax)
	pgplot.label(xlbl, ylbl, title)
	if mc > 1 then pgplot.setcolorindex(2) end
	pgplot.funt(fx, fy, tmin, tmax, true, nsamples)
	pgplot.setcolorindex()
end

-- polar: plot r = f(p), pmin <= p <= pmax
local PMIN = 0
local PMAX = 2 * PI
polar = function(f, pmin, pmax, nsamples, xlbl, ylbl, title)
	check_device()
	local pmin = check_arg(pmin, 2, 'number', PMIN)
	local pmax = check_arg(pmax, 3, 'number', PMAX)
	local nsamples = check_arg(nsamples, 4, 'number', STD_NSAMPLES)
	local xlbl = check_arg(xlbl, 5, 'string')
	local ylbl = check_arg(ylbl, 6, 'string')
	local title = check_arg(title, 7, 'string')
	local _, mc = pgplot.getcolorcap()
	if type(f) == 'function' then
		local x = function(p) return f(p) * cos(p) end
		local y = function(p) return f(p) * sin(p) end
		local xmin, xmax = find_fminmax(x, pmin, pmax, nsamples)
		local ymin, ymax = find_fminmax(y, pmin, pmax, nsamples)
		pgplot.environ(xmin, xmax, ymin, ymax)
		pgplot.label(xlbl, ylbl, title)
		if mc > 1 then pgplot.setcolorindex(2) end
		pgplot.funt(x, y, pmin, pmax, true, nsamples)
		pgplot.setcolorindex()
	elseif type(f) == 'table' then
		local n = getn(f)
		assert(n > 0, 'bad argument #1: empty table')
		local x = {}
		local y = {}
		-- get min and max for all functions
		assert(type(f[1]) == 'function', 'table value is not a function')
		x[1] = function(p) return f[1](p) * cos(p) end
		y[1] = function(p) return f[1](p) * sin(p) end
		local xmin, xmax = find_fminmax(x[1], pmin, pmax, nsamples)
		local ymin, ymax = find_fminmax(y[1], pmin, pmax, nsamples)
		for i = 2, n do
			assert(type(f[i]) == 'function', 'table value is not a function')
			x[i] = function(p) return f[i](p) * cos(p) end
			y[i] = function(p) return f[i](p) * sin(p) end
			local lxmin, lxmax = find_fminmax(x[i], pmin, pmax, nsamples)
			local lymin, lymax = find_fminmax(y[i], pmin, pmax, nsamples)
			if xmin > lxmin then xmin = lxmin end
			if xmax < lxmax then xmax = lxmax end
			if ymin > lymin then ymin = lymin end
			if ymax < lymax then ymax = lymax end
		end
		-- plot
		pgplot.environ(xmin, xmax, ymin, ymax)
		pgplot.label(xlbl, ylbl, title)
		pgplot.beginbuffer()
		for i = 1, n do
			if mc > 1 then
				pgplot.setcolorindex(mod(i - 1, mc - 1) + 2)
			end
			pgplot.funt(x[i], y[i], pmin, pmax, true, nsamples)
		end
		pgplot.setcolorindex()
		pgplot.endbuffer()
	else
		error('bad argument #1: table or function expected, got ' .. type(f))
	end
end


-- =====================================================================
-- Data related

-- scatter
local STD_SYMBOL = 4
data = function(x, y, symbol, xlbl, ylbl, title, sameplot)
	check_device()
	local sameplot = sameplot == nil and SAMEPLOT or sameplot
	local x = check_arg(x, 1, 'table')
	local n = getn(x)
	local y = y
	if type(y) ~= 'table' then
		-- switch
		y = x
		x = {}
		for i = 1, n do x[i] = i end
	else
		assert(n == getn(y), 'table sizes differ')
	end
	local symbol = check_arg(symbol, 3, 'number', STD_SYMBOL)
	local xlbl = check_arg(xlbl, 4, 'string')
	local ylbl = check_arg(ylbl, 5, 'string')
	local title = check_arg(title, 6, 'string')
	if not sameplot then -- set environment?
		local xmin = x[1]
		local xmax = x[1]
		local ymin = y[1]
		local ymax = y[1]
		for i = 2, n do
			if xmin > x[i] then xmin = x[i] end
			if xmax < x[i] then xmax = x[i] end
			if ymin > y[i] then ymin = y[i] end
			if ymax < y[i] then ymax = y[i] end
		end
		pgplot.environ(xmin, xmax, ymin, ymax)
	end
	pgplot.label(xlbl, ylbl, title)
	pgplot.point(x, y, symbol)
end

-- graph
graph = function(x, y, xlbl, ylbl, title, sameplot)
	check_device()
	local sameplot = sameplot == nil and SAMEPLOT or sameplot
	local x = check_arg(x, 1, 'table')
	local n = getn(x)
	local y = y
	if type(y) ~= 'table' then
		-- switch
		y = x
		x = {}
		for i = 1, n do x[i] = i end
	else
		assert(n == getn(y), 'table sizes differ')
	end
	local xlbl = check_arg(xlbl, 3, 'string')
	local ylbl = check_arg(ylbl, 4, 'string')
	local title = check_arg(title, 5, 'string')
	if not sameplot then -- set environment?
		local xmin = x[1]
		local xmax = x[1]
		local ymin = y[1]
		local ymax = y[1]
		for i = 2, n do
			if xmin > x[i] then xmin = x[i] end
			if xmax < x[i] then xmax = x[i] end
			if ymin > y[i] then ymin = y[i] end
			if ymax < y[i] then ymax = y[i] end
		end
		pgplot.environ(xmin, xmax, ymin, ymax)
	end
	pgplot.label(xlbl, ylbl, title)
	pgplot.line(x, y)
end


